/* 
 * Copyright (C) 2009 AG Projects
 * http://ag-projects.com
 */
#include <pjmedia/types.h>
#include <pjmedia/alaw_ulaw.h>
#include <pjmedia/errno.h>
#include <pjmedia/silencedet.h>
#include <pj/array.h>
#include <pj/assert.h>
#include <pj/lock.h>
#include <pj/log.h>
#include <pj/os.h>
#include <pj/pool.h>
#include <math.h>
#include <pjmedia.h>
#include <pjlib.h>

#include "echo_internal.h"

#define THIS_FILE                           "echo_ag.c"

/* Description:

   The AG Acoustic Echo Canceler is based on the NLMS-PW AEC coded by
   Andre Adrian[1]

   [1] http://www.andreadrian.de/intercom/index.html
*/

/* dB Values */
const double M0dB = 1.0f; 
const double M3dB = 0.71f;
const double M6dB = 0.50f;
const double M9dB = 0.35f;
const double M12dB = 0.25f;
const double M18dB = 0.125f;
const double M24dB = 0.063f;

/* dB values for 16bit PCM */
/* MxdB_PCM = 32767 * 10 ^(x / 20) */
const double M10dB_PCM = 10362.0f;
const double M20dB_PCM = 3277.0f;
const double M25dB_PCM = 1843.0f;
const double M30dB_PCM = 1026.0f;
const double M35dB_PCM = 583.0f;
const double M40dB_PCM = 328.0f;
const double M45dB_PCM = 184.0f;
const double M50dB_PCM = 104.0f;
const double M55dB_PCM = 58.0f;
const double M60dB_PCM = 33.0f;
const double M65dB_PCM = 18.0f;
const double M70dB_PCM = 10.0f;
const double M75dB_PCM = 6.0f;
const double M80dB_PCM = 3.0f;
const double M85dB_PCM = 2.0f;
const double M90dB_PCM = 1.0f;

const double MAXPCM = 32767.0f;

/* Ageing multiplier for LMS memory vector w */
const double LEAKY = 0.9999f;

/* Leaky ageing in samples. */
const int AGEING_TIME = 2 * 60 * 8;

/* scale factor*/
#define SCALE 10000

/* 
 * Exponential Smoothing or IIR Infinite Impulse Response Filter 
 */
typedef struct ag_filter_iir
{ 
    double x; 
    double a0;                     /* controls Transfer Frequency */
} ag_filter_iir; 

static void ag_filter_iir_init(ag_filter_iir *filter)
{
    filter->x = 0.0f;
    filter->a0 = 0.01f;           /* constant */
}

/* Highpass = Signal - Lowpass. Lowpass = Exponential Smoothing */
static double ag_filter_iir_highpass(ag_filter_iir *filter, double in)
{
    filter->x += filter->a0 * (in - filter->x); 
    return in - filter->x; 
}


/* 
 * 35 taps FIR Finite Impulse Response filter
 * Passband 150Hz to 4kHz for 8kHz sample rate, 300Hz to 8kHz for 16kHz
 * sample rate.
 * Coefficients calculated with
 * www.dsptutor.freeuk.com/KaiserFilterDesign/KaiserFilterDesign.html
 */

/* FIR Filter Size */
#define FIRF_SIZE 35

/* Kaiser Window FIR Filter, Filter type: High pass
 * Passband: 150.0 - 4000.0 Hz, Order: 34
 * Transition band: 34.0 Hz, Stopband attenuation: 10.0 dB
 */
const double firf_coefs[FIRF_SIZE + 1] = {
    -0.016165324, -0.017454365, -0.01871232, -0.019931411, 
    -0.021104068, -0.022222936, -0.02328091, -0.024271343, 
    -0.025187887, -0.02602462, -0.026776174, -0.027437767, 
    -0.028004972, -0.028474221, -0.028842418, -0.029107114, 
    -0.02926664, 0.8524841, -0.02926664, -0.029107114, 
    -0.028842418, -0.028474221, -0.028004972, -0.027437767, 
    -0.026776174, -0.02602462, -0.025187887, -0.024271343, 
    -0.02328091, -0.022222936, -0.021104068, -0.019931411, 
    -0.01871232, -0.017454365, -0.016165324, 0.0    
}; 

typedef struct ag_filter_fir
{
    double z[FIRF_SIZE + 1]; 
} ag_filter_fir; 

static void 
ag_filter_fir_init(ag_filter_fir *filter)
{
    pj_bzero(filter->z, (FIRF_SIZE + 1) * sizeof(filter->z[0]));
}

static double 
ag_filter_fir_highpass(ag_filter_fir *filter, double in)
{
    double *z = filter->z;         /* alias */
    double sum = 0.0;
    int i; 

    memmove(z + 1, z, FIRF_SIZE * sizeof(double)); 
    z[0] = in; 
    for (i = 0; i < FIRF_SIZE + 1; i++) {
        sum += firf_coefs[i] * z[i];
    }
    return sum; 
}

/* Recursive single pole IIR Infinite Impulse response High-pass
 * filter
 *
 * Reference: The Scientist and Engineer's Guide to Digital Processing
 *
 *      output[N] = A0 * input[N] + A1 * input[N-1] + B1 * output[N-1]
 *
 *      X  = exp(-2.0 * pi * Fc)
 *      A0 = (1 + X) / 2
 *      A1 = -(1 + X) / 2
 *      B1 = X
 *      Fc = cutoff freq / sample rate
 */

typedef struct ag_filter_rsiir
{
    double in0, out0; 
    double a0, a1, b1; 
} ag_filter_rsiir; 

static void 
ag_filter_rsiir_init(ag_filter_rsiir *filter, double fc)
{
    filter->in0 = 0.0f; 
    filter->out0 = 0.0f; 
    filter->b1 = expf(-2.0f * M_PI * fc); 
    filter->a0 = (1.0f + filter->b1) / 2.0f; 
    filter->a1 = - filter->a0;
}

static double 
ag_filter_rsiir_highpass(ag_filter_rsiir *filter, double in)
{
    double out;

    out = filter->a0 * in + 
        filter->a1 * filter->in0 + 
        filter->b1 * filter->out0;
    filter->in0 = in; 
    filter->out0 = out; 
    return out; 
}

/*
 * The AG echo canceler state
 */
typedef struct ag_aec_state 
{
    unsigned     clock_rate;          /* Clock rate.                               */
    pj_uint16_t  samples_per_frame;   /* Frame length in samples                   */
    pj_uint16_t  tail_ms;             /* Tail length in milliseconds               */
    pj_uint16_t  tail_len;            /* Tail length in samples.                   */
    pj_uint16_t  tail_ext;            /* Tail extenstion for memcopy optimization. */

    /* Time domain filters */
    ag_filter_iir mic_filter;	          /* DC-level remove Highpass */
    ag_filter_iir spk_filter;
    ag_filter_fir cutoff;                 /* 150Hz cut-off Highpass */
    ag_filter_rsiir pwhp_spk;             /* pre-whitening Highpass for speaker */
    ag_filter_rsiir pwhp_err;             /* pre-whitening Highpass for error */
    double gain;                           /* Mic signal amplify */
  
    /* Adrian soft decisi DTD (Double Talk Detector) */
    double mic_fast, spk_fast; 
    double mic_slow, spk_slow; 

    /* NLMS-pw */
    double *spk_samples;	       /* tap delayed loudspeaker signal */
    double *pw_samples;	       /* pre-whitening tap delayed signal */
    double *weights;	       /* tap weights */
    int tail_index;            /* optimize: less memory copies */
    double vecprod_pws_pws;    /* double to avoid loss of precision */
    double delta;	       /* noise floor to stabilize NLMS */

    int ageing; 

    pj_bool_t learning;		/* is true when learning is active */
    pj_bool_t cancelling; 		/* is true when cancelling is performed */

    /* AEC Algorithm tweak constants */
  
    /* minimum energy in xf. Range: M70dB_PCM to M50dB_PCM. Should be
     * equal to microphone ambient Noise level */
    double noise_floor;
  
    /* Double Talk Detector Speaker/Microphone Threshold. Range <=1
     * Large value (M0dB) is good for Single-Talk Echo cancellation,
     * small value (M12dB) is good for Doulbe-Talk AEC */
    double geigel_thold;

    /* for Non Linear Processor. Range >0 to 1. Large value (M0dB) is
     * good for Double-Talk, small value (M12dB) is good for
     * Single-Talk */
    double NLP_attenuation;

    double avg_spk; 

} ag_aec_state;

/* Util to display the error message for the specified error code  */
static int app_perror( const char *sender, const char *title, 
		       pj_status_t status)
{
    char errmsg[PJ_ERR_MSG_SIZE];

    PJ_UNUSED_ARG(sender);

    pj_strerror(status, errmsg, sizeof(errmsg));

    printf("%s: %s [code=%d]\n", title, errmsg, status);
    return 1;
}

/*
 * Create. 
 */
PJ_DEF(pj_status_t) 
ag_aec_create(pj_pool_t *pool,
              unsigned clock_rate,
              unsigned channel_count,
              unsigned samples_per_frame,
              unsigned tail_ms,
              unsigned options,
              void **p_state)
{
    ag_aec_state *ec;
    
    PJ_UNUSED_ARG(channel_count);
    PJ_UNUSED_ARG(options);

    ec = PJ_POOL_ZALLOC_T(pool, struct ag_aec_state);
    ec->clock_rate = clock_rate;
    ec->samples_per_frame = (pj_uint16_t)samples_per_frame;
    ec->tail_ms = (pj_uint16_t)tail_ms;
    ec->tail_len = (pj_uint16_t)(tail_ms * clock_rate / 1000); 
    /* set an extension for reducing memcopies */
    ec->tail_ext = ec->tail_len / 10; 

    /* algorithm tweak constants */
    ec->noise_floor = M55dB_PCM; 
    ec->geigel_thold = M6dB; 
    ec->NLP_attenuation = M6dB; 
    
    /* allocate sample arrays */
    ec->spk_samples = (double *) pj_pool_zalloc(pool, (ec->tail_len + ec->tail_ext) * sizeof(double));
    ec->pw_samples = (double *) pj_pool_zalloc(pool, (ec->tail_len + ec->tail_ext) * sizeof(double));
    ec->weights = (double *) pj_pool_zalloc(pool, ec->tail_len * sizeof(double));

    /* memcopy optimizer */
    ec->tail_index = ec->tail_ext;
    ag_aec_reset(ec);

    PJ_LOG(2, (THIS_FILE, "AG AEC created: tail_len = %d, tail_ext = %d", ec->tail_len, ec->tail_ext)); 
    
    *p_state = ec;
    return PJ_SUCCESS;
}


/*
 * Destroy. 
 */
PJ_DEF(pj_status_t) 
ag_aec_destroy(void *state)
{
    PJ_UNUSED_ARG(state);    

    PJ_LOG(2, (THIS_FILE, "AG AEC destroyed")); 
    return PJ_SUCCESS;
}


static void 
ag_aec_setambient(ag_aec_state *ec, double min_pws)
{
    /* subtract old delta */
    ec->vecprod_pws_pws -= ec->delta;
    /* compute new delta */
    ec->delta = (ec->tail_len - 1) * min_pws * min_pws; 
    /* add new delta */
    ec->vecprod_pws_pws += ec->delta; 
    ec->avg_spk = min_pws;
}

/*
 * Hard reset
 */
PJ_DEF(void) 
ag_aec_reset(void *state)
{
    ag_aec_state *ec = (ag_aec_state*) state;

    ec->delta = 0.0f;
    ec->vecprod_pws_pws = 0.0f;
    ag_aec_setambient(ec, ec->noise_floor);
    ec->mic_fast = ec->mic_slow = M75dB_PCM; 
    ec->spk_fast = ec->spk_slow = M80dB_PCM; 
    ec->gain = 1.0f; 

    /* init time domain filters */
    ag_filter_iir_init(&ec->mic_filter); 
    ag_filter_iir_init(&ec->spk_filter); 
    ag_filter_fir_init(&ec->cutoff); 
    ag_filter_rsiir_init(&ec->pwhp_spk, 2000.0f/ec->clock_rate); 
    ag_filter_rsiir_init(&ec->pwhp_err, 2000.0f/ec->clock_rate); 
    ec->learning = PJ_FALSE; 
    ec->cancelling = PJ_FALSE; 

    PJ_LOG(2, (THIS_FILE, "AG AEC reset ok, vecprod_pws_pws = %f, delta = %f, noise floor = %f", 
	       ec->vecprod_pws_pws, ec->delta, ec->noise_floor)); 
}

/*
 * Soft reset to force the EC to re-learn without having to discard
 * all rec and playback history.
 */
PJ_DEF(void) 
ag_aec_soft_reset(void *state)
{
    PJ_LOG(4,(THIS_FILE, "AG AEC soft reset. Re-learning.."));
}

/*
 *  Adrian soft decision DTD (Dual Average Near-End to Far-End signal
 * Ratio DTD) This algorithm uses exponential smoothing with differnt
 * ageing parameters to get fast and slow near-end and far-end signal
 * averages. The ratio of NFRs term (mic_fast / spk_fast) / (mic_slow / spk_slow)
 * is used to compute the stepsize A ratio value of 2.5 is mapped to
 * stepsize 0, a ratio of 0 is mapped to 1.0 with a limited linear
 * function.
 */

/* left point. X is ratio, Y is stepsize */
const double STEPX1 = 1.0; 
const double STEPY1 = 1.0;
/* right point. STEPX2=2.0 is good double talk, 3.0 is good single talk. */
const double STEPX2 = 2; 
const double STEPY2 = 0;
const double ALPHAFAST = 1.0f / 100.0f;
const double ALPHASLOW = 1.0f / 20000.0f;

#define TALK_LIMIT M60dB_PCM

static double
ag_aec_varstepsize(ag_aec_state *ec)
{
    double stepsize; 
    double M = (STEPY2 - STEPY1) / (STEPX2 - STEPX1); 
    /* ratio of NFRs */
    double ratio = (ec->mic_fast * ec->spk_slow) / (ec->mic_slow * ec->spk_fast);

    /* begrenzte lineare Kennlinie */
    if (ratio < STEPX1) {
        stepsize = STEPY1; 
    }
    else if (ratio > STEPX2) {
        stepsize = STEPY2;
    }
    else {
        /* Punktrichtungsform einer Geraden */
        stepsize = M * (ratio - STEPX1) + STEPY1;
    }

    return stepsize; 

}

/* Double-Talk Detector
 *
 * in mic_sample: microphone sample (PCM as floating point value) in
 * spk_sample: loudspeaker sample (PCM as floating point value)
 * return: from 0 for doubletalk to 1.0 for single talk
 *
 */
static double 
ag_aec_dtd(ag_aec_state *ec, double mic_sample, double spk_sample)
{
    double stepsize; 

    /* fast near-end and far-end average */
    ec->mic_fast += ALPHAFAST * (fabsf(mic_sample) - ec->mic_fast);
    ec->spk_fast += ALPHAFAST * (fabsf(spk_sample) - ec->spk_fast);
  
    /* slow near-end and far-end average */
    ec->mic_slow += ALPHASLOW * (fabsf(mic_sample) - ec->mic_slow);
    ec->spk_slow += ALPHASLOW * (fabsf(spk_sample) - ec->spk_slow);

    //PJ_LOG(2, (THIS_FILE, "Spk:%f, Mic: %f, Limit: %f", ec->spk_fast, ec->mic_fast, M70dB_PCM)); 
    ec->learning = PJ_TRUE; 
    ec->cancelling = PJ_TRUE; 
    stepsize = ag_aec_varstepsize(ec); 

    if (ec->spk_fast < TALK_LIMIT || ec->mic_fast < TALK_LIMIT) {
	ec->learning = PJ_FALSE; 
	ec->cancelling = PJ_TRUE; 
    }
    
/*     if ((ec->spk_fast < TALK_LIMIT) && (ec->mic_fast < TALK_LIMIT)) { */
/* 	//PJ_LOG(2, (THIS_FILE, "Idle line")); */
/* 	/\* idle line, do nothing *\/ */
/* 	ec->learning = PJ_FALSE; */
/* 	ec->cancelling = PJ_FALSE; */
/* 	pj_bzero(ec->weights, ec->tail_len); */
/* 	stepsize = ag_aec_varstepsize(ec); */
/*     } else if ((ec->spk_fast >= TALK_LIMIT) && (ec->mic_fast < TALK_LIMIT)) { */
/* 	//PJ_LOG(2, (THIS_FILE, "Far end active, Near end idle")); */
/* 	//PJ_LOG(2, (THIS_FILE, "Spk:%f, Mic: %f, Limit: %f", ec->spk_fast, ec->mic_fast, limit)); */
/* 	/\* FarEnd active, NearEnd idle *\/ */
/* 	ec->learning = PJ_TRUE; */
/* 	ec->cancelling = PJ_TRUE; */
/* 	stepsize = ag_aec_varstepsize(ec); */
/*     } else if ((ec->spk_fast < TALK_LIMIT) && (ec->mic_fast >= TALK_LIMIT)) { */
/* 	//PJ_LOG(2, (THIS_FILE, "Far end idle, Near end active")); */
/* 	ec->learning = PJ_TRUE; */
/* 	ec->cancelling = PJ_TRUE; */
/* 	stepsize = ag_aec_varstepsize(ec); */
/*     } else { */
/* 	//PJ_LOG(2, (THIS_FILE, "DTD")); */
/* 	//return 0.0; */
/* 	/\* double talk *\/ */
/* 	ec->learning = PJ_TRUE; */
/* 	ec->cancelling = PJ_TRUE; */
/* 	stepsize = ag_aec_varstepsize(ec); */
/*     } */

    return stepsize; 
}

/* 
 * The spk_fast signal is used to charge the ageing timer to Thold.
 * When ageing expires (no Spk signal for some time) the vector w is
 * erased. This is my implementation of Leaky NLMS.
 */
static void 
ag_aec_leaky(ag_aec_state *ec)
{
    if (ec->spk_fast >= TALK_LIMIT) {
        /* vector weight is valid for ageing Thold time */
        ec->ageing = AGEING_TIME; 
    }
    else
        {
            if (ec->ageing > 1) {
                ec->ageing--; 
            }
            else if (1 == ec->ageing) {
		ec->ageing--; 
                /* Leaky NLMS is to erase vector w when ageing expires */
                pj_bzero(ec->weights, ec->tail_len); 
            }
        }
}

/* vector dot product */
static double 
ag_vec_prod(double *a, double *b, int size)
{
    double sum = 0.0; 
    int i; 

    for (i = 0; i < size; i++) {                                           
        sum += a[i] * b[i]; 
    }

    return sum; 
}

/* Normalized Least Mean Square Algorithm pre-whitening (NLMS-pw) The
 * LMS algorithm was developed by Bernard Widrow book: Haykin,
 * Adaptive Filter Theory, 4. edition, Prentice Hall, 2002
 *
 * in mic_sample: microphone sample (16bit PCM value) in spk_sample:
 * loudspeaker sample (16bit PCM value) in stepsize: NLMS adaptation
 * variable return: echo cancelled microphone sample
 */
static double 
ag_aec_nlms_pw(ag_aec_state *ec, double mic_sample, double spk_sample, double stepsize)
{
    int j = ec->tail_index; 
    
    ec->spk_samples[j] = spk_sample; 
    ec->pw_samples[j] = ag_filter_rsiir_highpass(&ec->pwhp_spk, spk_sample); 
  
    /* optimize: iterative vec_prod(ec->pw_samples, ec->pw_samples) */
    ec->vecprod_pws_pws += (ec->pw_samples[j] * ec->pw_samples[j]);
    ec->vecprod_pws_pws -= (ec->pw_samples[j + ec->tail_len - 1] * ec->pw_samples[j + ec->tail_len - 1]);

    /* average speaker signal */
    //ec->avg_spk += (ec->pw_samples[j] - ec->pw_samples[j + ec->tail_len - 1]) / (ec->tail_len - 1);

    /* compute echo cancelled signal (mic signal - estimated mic signal from
     * spk signal)
     */
    double ec_signal = mic_sample; 
    double pw_ec = ec_signal; 
    if (ec->cancelling == PJ_TRUE) {
        ec_signal -= ag_vec_prod(ec->weights, ec->spk_samples + j, ec->tail_len); 
	/* pre-whitening of echo cancelled signal */
	pw_ec = ag_filter_rsiir_highpass(&ec->pwhp_err, ec_signal); 
    }

    //PJ_LOG(2, (THIS_FILE, "Mic signal: %f; EC signal: %f; Pw-echo-cancelled: %f; ageing: %d", mic_sample, ec_signal, pw_ec, ec->ageing));
    
    if (ec->learning == PJ_TRUE) {                          
        /* calculate variable step size */
	//double mikro_ef = ec->avg_spk * stepsize / ec->vecprod_pws_pws; 
	//double vecprod_pws_pws = ag_vec_prod(ec->pw_samples + j, ec->pw_samples + j, ec->tail_len); 
	//PJ_LOG(2, (THIS_FILE, "vecprod_pws_pws: %lf default: %lf", vecprod_pws_pws, ec->vecprod_pws_pws)); 
	double mikro_ef = stepsize * pw_ec / ec->vecprod_pws_pws; 
	
        /* update tap weights (filter learning) */
	//PJ_LOG(2, (THIS_FILE, "filter learning: %f %lf %f %f %f", ec->avg_spk, SCALE * stepsize / ec->vecprod_pws_pws, pw_ec, ec->vecprod_pws_pws, 1 / ec->vecprod_pws_pws));
	int i; 
        for (i = 0; i < ec->tail_len; i++) {
	    ec->weights[i] += mikro_ef * ec->pw_samples[i + j];
        }
    }

    ec->tail_index--; 
    if (ec->tail_index < 0) {
        /* optimize: decrease number of memory copies */
        ec->tail_index = ec->tail_ext; 
	j = ec->tail_ext; 
        memmove(ec->spk_samples + j + 1, ec->spk_samples, 
                (ec->tail_len - 1) * sizeof(double));
        memmove(ec->pw_samples + j + 1, ec->pw_samples, 
                (ec->tail_len - 1) * sizeof(double));
    }
  
    /* saturation */
    double sat = ec_signal; 
    if (ec_signal > MAXPCM) {
        sat = MAXPCM;  
    }
    else if (ec_signal < -MAXPCM) {
        sat = -MAXPCM; 
    }
    
    return sat; 
}

/* Acoustic Echo Cancellation and Suppression of one sample in mic:
 * microphone signal with echo in spk: loudspeaker signal return: echo
 * cancelled microphone signal
 */
static int 
ag_aec_compute(ag_aec_state *ec, int mic, int spk)
{
    double mic_sample = (double) mic; 
    double spk_sample = (double) spk; 
  
    //PJ_LOG(2, (THIS_FILE, "Initial mic signal: %d %f", mic, mic_sample));

    /* Mic Highpass Filter - to remove DC */
    mic_sample = ag_filter_iir_highpass(&ec->mic_filter, mic_sample);
  
    //PJ_LOG(2, (THIS_FILE, "after iir_highpass: %f", mic_sample)); 
    
    /* Mic Highpass Filter - cut-off below 300Hz */
    mic_sample = ag_filter_fir_highpass(&ec->cutoff, mic_sample); 

    //PJ_LOG(2, (THIS_FILE, "after fir_highpass: %f", mic_sample)); 

    /* Amplify, for e.g. Soundcards with -6dB max. volume */
    mic_sample *= ec->gain; 

    //PJ_LOG(2, (THIS_FILE, "after amplify: %f", mic_sample)); 

    /* Spk Highpass Filter - to remove DC */
    spk_sample = ag_filter_iir_highpass(&ec->spk_filter, spk_sample); 

    //PJ_LOG(2, (THIS_FILE, "spk after iir_highpass %f", spk_sample));

    /* Double Talk Detector */
    double stepsize = ag_aec_dtd(ec, mic_sample, spk_sample);
    //PJ_LOG(2, (THIS_FILE, "stepsize= %f", stepsize));    

    /* Leaky (ageing of vector weights) */
    ag_aec_leaky(ec); 

    /* Acoustic Echo Cancellation */
    mic_sample = ag_aec_nlms_pw(ec, mic_sample, spk_sample, stepsize);

    //PJ_LOG(2, (THIS_FILE, "after nlms: %f (int) %d", mic_sample, (int)roundf(mic_sample))); 

    return (int) roundf(mic_sample); 
}

#define SAMPLE_CHARS 32

static void
log_frame(ag_aec_state *ec, pj_int16_t *frm)
{						
    int i, bytes; 
    char *buf, *pbuf;

    buf = malloc(SAMPLE_CHARS * ec->samples_per_frame);

    buf[0] = 0;
    pbuf = buf; 
    for (i = 0; i < ec->samples_per_frame; i++) {
	bytes = sprintf(pbuf, "%10d", frm[i]);
	pbuf += bytes;
    }
    PJ_LOG(2, (THIS_FILE, buf));

    free(buf); 
}

static void 
log_samples(ag_aec_state *ec, double *v)
{
    int i, bytes; 
    char *buf, *pbuf;
    
    buf = malloc(SAMPLE_CHARS * (ec->tail_len + ec->tail_ext));

    buf[0] = 0;
    pbuf = buf; 
    for (i = ec->tail_index; i < ec->tail_len + ec->tail_index; i++) {
	bytes = sprintf(pbuf, "%10.2f", v[i]);
	pbuf += bytes;
    }
    PJ_LOG(2, (THIS_FILE, buf));

    free(buf); 
}


static void 
log_weights(ag_aec_state *ec)
{
    int i, bytes; 
    char *buf, *pbuf;
    
    buf = malloc(SAMPLE_CHARS * ec->tail_len);

    buf[0] = 0;
    pbuf = buf; 
    for (i = 0; i < ec->tail_len; i++) {
	bytes = sprintf(pbuf, "%10f", ec->weights[i]);
	pbuf += bytes;
    }
    PJ_LOG(2, (THIS_FILE, buf));

    free(buf); 
}

/* 
 * Perform echo cancellation.
 */
PJ_DEF(pj_status_t) 
ag_aec_cancel_echo(void *state,
                   pj_int16_t *rec_frm,
                   const pj_int16_t *play_frm,
                   unsigned options,
                   void *reserved)
{
    ag_aec_state *ec = (ag_aec_state*) state;
    int unsigned i; 

    PJ_UNUSED_ARG(options); 
    PJ_UNUSED_ARG(reserved); 

    //PJ_LOG(2, (THIS_FILE, "Before AEC processing..."));
    //log_frame(ec, rec_frm); 

    /* perform echo cancellation on each sample from the frame */
    for (i = 0; i < ec->samples_per_frame; i++) {
	rec_frm[i] = ag_aec_compute(ec, rec_frm[i], play_frm[i]); 
    }
    
    //PJ_LOG(2, (THIS_FILE, "Samples, tail_index = %d :", ec->tail_index));
    //log_samples(ec, ec->spk_samples);
    //PJ_LOG(2, (THIS_FILE, "After AEC processing..."));
    //log_frame(ec, rec_frm);     
    
    //PJ_LOG(2, (THIS_FILE, "Weights:"));
    //log_weights(ec); 
    //PJ_LOG(2, (THIS_FILE, "Samples:"));
    //log_pwsamples(ec); 

    return PJ_SUCCESS;
}

